#!/usr/bin/env python3
# ----------------------------------------------------------------------
# Copyright (C) 2025 Tristan Hoellinger
# Distributed under the GNU General Public License v3.0 (GPLv3).
# See the LICENSE file in the root directory for details.
# SPDX-License-Identifier: GPL-3.0-or-later
# ----------------------------------------------------------------------

__author__ = "Tristan Hoellinger"
__version__ = "0.1.0"
__date__ = "2025"
__license__ = "GPLv3"

"""Global parameters for this project."""

import os
from pathlib import Path
from typing import cast
import numpy as np

WHICH_SPECTRUM = "eh"  # available options are "eh" and "class"

ROOT_PATH = cast(str, os.getenv("WIP3M_ROOT_PATH"))
if ROOT_PATH is None:
    raise EnvironmentError("Please set the 'WIP3M_ROOT_PATH' environment variable.")

OUTPUT_PATH = cast(str, os.getenv("WIP3M_OUTPUT_PATH"))
if OUTPUT_PATH is None:
    raise EnvironmentError("Please set the 'WIP3M_OUTPUT_PATH' environment variable.")

# Default verbose level
# 0: errors only, 1: info, 2: warnings+, 3: all diagnostics, 4+: debug
DEFAULT_VERBOSE_LEVEL = 2

# Baseline seeds for reproducibility
BASELINE_SEEDNORM = 100050599
BASELINE_SEEDNOISE = 200050599
BASELINE_SEEDPHASE = 300050599

# Fiducial cosmological parameters
h_planck = 0.6766
Omega_b_planck = 0.02242 / h_planck**2
Omega_m_planck = 0.3111
nS_planck = 0.9665
sigma8_planck = 0.8102

TAU_REIO = 0.066

planck_mean = np.array([h_planck, Omega_b_planck, Omega_m_planck, nS_planck, sigma8_planck])
planck_cov = np.diag(np.array([0.0042, 0.00030, 0.0056, 0.0038, 0.0060]) ** 2)

# Mapping from cosmological parameter names to corresponding indices
cosmo_params_names = [r"$h$", r"$\Omega_b$", r"$\Omega_m$", r"$n_S$", r"$\sigma_8$"]
cosmo_params_name_to_idx = {"h": 0, "Omega_b": 1, "Omega_m": 2, "n_s": 3, "sigma8": 4}

# Minimum k value used in the normalisation of the summaries
MIN_K_NORMALISATION = 4e-2

params_planck_kmax_missing = {
    "h": h_planck,
    "Omega_r": 0.0,
    "Omega_q": 1.0 - Omega_m_planck,
    "Omega_b": Omega_b_planck,
    "Omega_m": Omega_m_planck,
    "m_ncdm": 0.0,
    "Omega_k": 0.0,
    "tau_reio": TAU_REIO,
    "n_s": nS_planck,
    "sigma8": sigma8_planck,
    "w0_fld": -1.0,
    "wa_fld": 0.0,
    "WhichSpectrum": WHICH_SPECTRUM,
}

Omega_b_CONCEPT = 0.049
Omega_cdm_CONCEPT = 0.27
Omega_m_CONCEPT = Omega_b_CONCEPT + Omega_cdm_CONCEPT
params_CONCEPT_kmax_missing = {
    "h": 0.67,
    "Omega_r": 0.0,
    "Omega_q": 1.0 - Omega_m_CONCEPT,
    "Omega_b": Omega_b_CONCEPT,
    "Omega_m": Omega_m_CONCEPT,
    "m_ncdm": 0.0,
    "Omega_k": 0.0,
    "tau_reio": TAU_REIO,
    "n_s": 0.96,
    "sigma8": 0.8102,
    "w0_fld": -1.0,
    "wa_fld": 0.0,
    "WhichSpectrum": WHICH_SPECTRUM,
}


def z2a(z):
    return 1.0 / (1 + z)


def cosmo_small_to_full_dict(cosmo_min):
    """Return a full cosmology dictionary from a minimal one.

    Parameters
    ----------
    cosmo_min : dict
        Minimal cosmology dictionary.

    Returns
    -------
    cosmo_full : dict
        Full cosmology dictionary.

    """
    cosmo_full = {
        "h": cosmo_min["h"],
        "Omega_r": 0.0,
        "Omega_q": 1 - cosmo_min["Omega_m"],
        "Omega_b": cosmo_min["Omega_b"],
        "Omega_m": cosmo_min["Omega_m"],
        "m_ncdm": 0.0,
        "Omega_k": 0.0,
        "tau_reio": TAU_REIO,
        "n_s": cosmo_min["n_s"],
        "sigma8": cosmo_min["sigma8"],
        "w0_fld": -1.0,
        "wa_fld": 0.0,
        "k_max": cosmo_min["k_max"],
        "WhichSpectrum": cosmo_min["WhichSpectrum"],
    }
    return cosmo_full

# Default parameters time step limiters
# Mostly for development and testing purposes.
# For real P3M+COLA runs starting at redshift z=19, only
# DEFAULT_FAC_DYN_CUSTOM_COLA, DEFAULT_FAC_H_CUSTOM_COLA,
# DEFAULT_FAC_P3M_FIT, DEFAULT_DA_MAX_LATE_CUSTOM and
# P3M_FIT_COEFFS_DEFAULT_{1,2}NP are relevant.
DEFAULT_FAC_DYN_CUSTOM = 0.0154
DEFAULT_FAC_DYN_CUSTOM_COLA = 0.03
DEFAULT_DA_MAX_EARLY_CUSTOM = 0.0013
DEFAULT_FAC_H_CUSTOM = 0.03
DEFAULT_FAC_H_CUSTOM_COLA = 0.05
DEFAULT_FAC_BEND = 0.3125
DEFAULT_SUB_BEND1 = 0.012
DEFAULT_SUB_BEND2 = 0.007
DEFAULT_SUB_BEND1_COLA = 0.012
DEFAULT_SUB_BEND2_COLA = 0.014
DEFAULT_FAC_P3M_FIT = 1.0
DEFAULT_DA_MAX_LATE_CUSTOM = 0.02

# V1:
# P3M_FIT_COEFFS_DEFAULT_2NP = [-4.41928363, 0.24380666, 1.92034984, 1.21626721, 0.81694801, 0.20590599]  # fitted P3M limiter for L=2Np
# P3M_FIT_COEFFS_DEFAULT_NP = [-5.14695921, 0.35234941, 2.08397701, 1.2447759, 0.61695249, 0.11401126]  # fitted P3M limiter for L=Np

# V2:
P3M_FIT_COEFFS_DEFAULT_2NP = [
    -2.421323,
    0.45965703,
    2.52817976,
    1.58540987,
    0.81381042,
    0.17553455,
]  # fitted P3M limiter for L=2Np based on COLAr6/V3cola_Pf_L1024_N1024_Np512
P3M_FIT_COEFFS_DEFAULT_NP = [
    -3.19544528,
    -0.01948223,
    0.5180241,
    -1.08372045,
    -0.71808255,
    -0.14370312,
]  # fitted P3M limiter for L=Np based on COLAr6/V3cola_Pf_L512_N1024_Np512
P3M_FIT_COEFFS_DEFAULT_05NP = [
    -3.74659557,
    0.39041298,
    1.90821078,
    1.10022755,
    0.5597775,
    0.09604354,
]  # fitted P3M limiter for L=Np based on COLAr7/CV2_cola_Pf_L80_N160_Np160

# pyright: reportUnsupportedDunderAll=false
__all__ = [k for k in globals() if not k.startswith("_")]
